open Types;;
open Set;;
open Ineq;;

let fv_typ t =
 let processed = ref poly_empty in

 let rec loop p u =
  let u = repr u in
  if List.memq u !processed.tvar then p
  else begin
   processed := { !processed with tvar = u::(!processed.tvar) };
   match u.data with
      TVar -> { p with tvar = setq_add u p.tvar }
    | FunT (u1, u2, u3) -> loop (loopd (loop p u1) u2) u3
    | RecordT u1 -> loopr p u1
    | ConstrT (_, u1) -> loop p u1
    | _ -> p
  end

 and loopr p u =
  let u = repr u in
  if List.memq u !processed.rvar then p
  else begin
   processed := { !processed with rvar = u::(!processed.rvar) };
   match u.data with
      RVar _ -> { p with rvar = setq_add u p.rvar }
    | LabelT (_, u1) -> loop p u1
    | NodeT (u1, u2) -> loopr (loopr p u1) u2
    | EmptyT -> p
  end

 and loopd p u =
  let u = repr u in
  if List.memq u !processed.dvar then p
  else begin
   processed := { !processed with dvar = u::(!processed.dvar) };
   match u.data with
      DVar _ -> { p with dvar = setq_union (comp_connexe u) p.dvar }
    | Inf (u1, u2) -> loopd (loopd p u1) u2
    | _ -> p
  end

 in loop poly_empty t;;

let fv_scheme ((p, t) : scheme) = poly_minus (fv_typ t) p;;

let fv_envT (e : envT) =
 List.fold_left (fun p (_,s) -> poly_union (fv_scheme s) p) poly_empty e;;

let fv2_deg d =
 let processed = ref poly_empty in

 let rec loop l u =
  let u = repr u in
  if List.memq u !processed.tvar then l
  else begin
   processed := { !processed with tvar = u::(!processed.tvar) };
   match u.data with
      FunT (u1, u2, u3) -> loop (loopd (loop l u1) u2) u3
    | RecordT u1 -> loopr l u1
    | ConstrT (_, u1) -> loop l u1
    | _ -> l
  end

 and loopr l u =
  let u = repr u in
  if List.memq u !processed.rvar then l
  else begin
   processed := { !processed with rvar = u::(!processed.rvar) };
   match u.data with
      LabelT (_, u1) -> loop l u1
    | NodeT (u1, u2) -> loopr (loopr l u1) u2
    | _ -> l
  end

 and loopd l u =
  let u = repr u in
  if List.memq u !processed.dvar then l
  else begin
   processed := { !processed with dvar = u::(!processed.dvar) };
   match u.data with
      DVar _ -> setq_add u l
    | Inf (u1, u2) -> loopd (loopd l u1) u2
    | _ -> l
  end

 in loop setq_empty d;;
