{
open Common;;
open Parser;;
exception Eof;;

let elim_quotes s = String.sub s 1 (String.length s - 2);;
}

let comm1 = "//" [^ '\n']* '\n'
let comm2 = "/*" ([^ '*'] | '*'[^ '/'])* "*/"
let delim = [' ' '\t' '\n']
let bl = delim+
let chiffre = ['0'-'9']
let lettre = ['A'-'Z' 'a'-'z']
let id = (lettre | '_') (lettre | chiffre | '_' | '\'')*
let tvarid = '\'' ['a'-'z'] (lettre | chiffre | '_' | '\'')*
let rvarid = '\'' ['A'-'Z'] (lettre | chiffre | '_' | '\'')*
let nombre = ('-')? chiffre+
let string = '"' [^ '"']* '"'

rule token = parse
   bl | comm1 | comm2      { token lexbuf }
 | "let"                   { LET }
 | "rec"                   { REC }
 | "and"                   { AND }
 | "type"                  { TYPE }
 | "unsafe"                { UNSAFE }
 | "use"                   { USE }
 | '='                     { EQ }
 | "in"                    { IN }
 | "fun"                   { FUN }
 | "->"                    { TO }
 | "if"                    { IF }
 | "then"                  { THEN }
 | "else"                  { ELSE }
 | "mixin"                 { MIXIN }
 | "end"                   { END }
 | "var"                   { VAR }
 | "cst"                   { CST }
 | "meth"                  { METH }
 | "inherit"               { INHERIT }
 | "as"                    { AS }
 | "without"               { WITHOUT }
 | "rename"                { RENAME }
 | '#'                     { DIESE }
 | '('                     { LPAR }
 | ')'                     { RPAR }
 | '{'                     { LACC }
 | '}'                     { RACC }
 | '['                     { LBRACKET }
 | ']'                     { RBRACKET }
 | '.'                     { DOT }
 | '\\'                    { BACK }
 | ','                     { COMMA }
 | "<-"                    { OVERRIDE }
 | ";;"                    { ENDINSTR }
 | '!'                     { BANG }
 | ';'                     { SEMICOLON }
 | ':'                     { COLON }
 | "::"                    { CONS }
 | ":="                    { ASSIGN }
 | "&&"                    { ANDB }
 | "||"                    { OR }
 | "@"                     { LISTCONCAT }
 | "^"                     { STRINGCONCAT }
 | '+' | '-'               { INFIXPLUSMOINS(Lexing.lexeme lexbuf) }
 | '*' | '/' | '%'         { INFIXFOISDIV(Lexing.lexeme lexbuf) }
 | '<' | "<=" | '>' | ">="
       |  "==" | "<>"      { INFIXCOMP(Lexing.lexeme lexbuf) }
 | nombre                  { INT(int_of_string(Lexing.lexeme lexbuf)) }
 | id                      { ID(Lexing.lexeme lexbuf) }
 | tvarid                  { TVARID(Lexing.lexeme lexbuf) }
 | rvarid                  { RVARID(Lexing.lexeme lexbuf) }
 | string                  { STRING(elim_quotes (Lexing.lexeme lexbuf)) }
 | eof                     { raise Eof }
 | _                       { runtime_msg "Lexical error" }
