%{
open Types;;
open Syntax;;

let param_list = List.fold_right (fun (x,topt) l -> Fun (x, topt, l));;

let appl_list_to_fun l e =
 let rec loop l = function
    Ident lab -> RCons(REmpty, lab, param_list l e)
  | Appl(r, Ident x) -> loop ((x,None)::l) r
  | Appl(r, TypeConstr (Ident x, t)) -> loop ((x,Some t)::l) r
  | _ -> raise Parsing.Parse_error
 in loop [] l;;

let fresh_var =
 let compt = ref (-1) in
 fun () -> incr compt; "#" ^ string_of_int !compt;;

let mixin_inherit t m =
 let g = fresh_var () and s = fresh_var () in
 Fun (g, None, Fun (s, None, Appl (Appl (m, Appl (t, Ident g)), Ident s)));;

let mixin_var t l n =
 let g = fresh_var () and s = fresh_var () in
 Fun (g, None, Fun (s, None, RCons (Appl (Appl (t, Ident g), Ident s),
			            l, Appl (Ident "ref", n))));;

let mixin_cst t l n =
 let g = fresh_var () and s = fresh_var () in
 Fun (g, None, Fun (s, None, RCons (Appl (Appl (t, Ident g), Ident s),
				    l, n)));;

let mixin_meth t l (y,ty) (x,tx) n =
 let g = fresh_var () and s = fresh_var ()
 and u = fresh_var () and z = fresh_var () in
 Fun (g, None, Fun (s, None, Let ([(z, Appl (Appl (t, Ident g), Ident s))],
    RCons (Ident z,
           l, Fun (u, Some unit_cell,
		   Appl (Appl (Fun (y, ty, Fun (x, tx, n)),
			       Ident z), Ident s))))));;

let mixin_meth_over t l (y,ty) (x,tx) n =
 let g = fresh_var () and s = fresh_var ()
 and u = fresh_var () and z = fresh_var () in
 Fun (g, None, Fun (s, None, Let ([(z, Appl (Appl (t, Ident g), Ident s))],
   RCons (RRes (Ident z, l),
	  l, Fun (u, Some unit_cell,
		  Appl (Appl (Fun (y, ty, Fun (x, tx, n)),
                              Ident z), Ident s))))));;

let mixin_without t l =
 let g = fresh_var () and s = fresh_var () in
 Fun (g, None, Fun (s, None, RRes (Appl (Appl (t, Ident g), Ident s), l)));;

let mixin_rename t l l' =
 let g = fresh_var () and s = fresh_var () and x = fresh_var () in
 Fun (g, None, Fun (s, None, Let ([(x, Appl (Appl (t, Ident g), Ident s))],
                      RCons (RRes (Ident x, l), l', RSel (Ident x, l)))));;

let empty () =
 let g = fresh_var () and s = fresh_var () in
 Fun (g, None, Fun (s, None, Appl (Ident g, Ident s)));;

let typ_of_id = function
   "int" -> int_cell
 | "bool" -> bool_cell
 | "unit" -> unit_cell
 | "string" -> string_cell
 | _ -> raise Parsing.Parse_error;;

let (reset_typid, typ_of_tvarid, typ_of_rvarid) =
 let l_tvar = ref [] and l_rvar = ref [] in
 (fun () -> l_tvar := []; l_rvar := []),
 (fun a -> if List.mem_assoc a !l_tvar
           then List.assoc a !l_tvar
           else let t = tvar_cell () in
	        l_tvar := (a,t)::(!l_tvar);
	      	t),
(* erreur si on reutilise 'A avec des contraintes differentes !!! *)
 (fun a constr -> if List.mem_assoc a !l_rvar
	          then List.assoc a !l_rvar
        	  else let t = rvar_cell constr in
	               l_rvar := (a,t)::(!l_rvar);
	      	       t);;
%}

%token <int> INT
%token <string> ID TVARID RVARID STRING
%token LET REC AND EQ IN FUN TO IF THEN ELSE TYPE UNSAFE USE
%token MIXIN END VAR CST METH WITHOUT RENAME
%token INHERIT AS DIESE
%token LPAR RPAR LACC RACC LBRACKET RBRACKET
%token DOT BACK COMMA OVERRIDE ENDINSTR BANG SEMICOLON COLON
%token <string> INFIXPLUSMOINS INFIXFOISDIV INFIXCOMP
%token ANDB OR ASSIGN CONS LISTCONCAT STRINGCONCAT

%nonassoc IN
%right TO
%right SEMICOLON
%nonassoc ELSE
%right ASSIGN
%left OR
%left ANDB
%nonassoc INFIXCOMP
%right LISTCONCAT STRINGCONCAT
%right CONS
%left INFIXPLUSMOINS
%left INFIXFOISDIV
%right BANG
%left DOT DIESE
%left BACK
%nonassoc ID

%start main
%type <Syntax.instr> main

%%

main:
    instr ENDINSTR                    { reset_typid(); $1 }
;

instr:
    LET let_binding                   { Def $2 }
  | LET REC let_binding               { DefRec $3 }
  | expr                              { Expr $1 }
  | DIESE TYPE expr                   { Type $3 }
  | DIESE UNSAFE expr                 { Unsafe $3 }
  | DIESE USE STRING                  { Use $3 }
;

expr:
    LET let_binding IN expr           { Let($2,$4) }
  | LET REC let_binding IN expr       { LetRec($3,$5) }
  | FUN id params TO expr             { Fun(fst $2,snd $2,param_list $3 $5) }
  | IF expr THEN expr ELSE expr       { If($2,$4,$6) }
  | MIXIN mixin END                   { $2 }
  | expr INFIXPLUSMOINS expr          { Appl(Appl(Ident $2,$1),$3) }
  | expr INFIXFOISDIV expr            { Appl(Appl(Ident $2,$1),$3) }
  | expr ASSIGN expr                  { Appl(Appl(Ident ":=",$1),$3) }
  | expr INFIXCOMP expr               { Appl(Appl(Ident $2,$1),$3) }
  | expr ANDB expr                    { Appl(Appl(Ident "&&",$1),$3) }
  | expr OR expr                      { Appl(Appl(Ident "||",$1),$3) }
  | expr CONS expr                    { Appl(Appl(Ident "::",$1),$3) }
  | expr SEMICOLON expr               { Appl(Appl(Ident ";",$1),$3) }
  | expr LISTCONCAT expr              { Appl(Appl(Ident "@", $1), $3) }
  | expr STRINGCONCAT expr            { Appl(Appl(Ident "^", $1), $3) }
  | appl                              { $1 }
;

let_binding:
    ID params EQ expr                 { [($1,param_list $2 $4)] }
  | ID params EQ expr AND let_binding { ($1,param_list $2 $4)::$6 }
;

appl:
    appl term                         { Appl($1,$2) }
  | term                              { $1 }
;

term:
    ID                                { Ident $1 }
  | INT                               { IntE $1 }
  | STRING                            { StringE $1 }
  | LACC record RACC                  { $2 }
  | LPAR expr RPAR                    { $2 }
  | LPAR expr COLON typexpr RPAR      { TypeConstr($2,$4) }
  | LPAR op RPAR                      { Ident $2 }
  | LBRACKET list RBRACKET            { $2 }
  | term DOT ID                       { RSel($1,$3) }
  | term DIESE ID                     { Appl(RSel($1,$3),Ident "()") }
  | term BACK ID                      { RRes($1,$3) }
  | BANG term                         { Appl(Ident "!",$2) }
;

op:
    INFIXPLUSMOINS                    { $1 }
  | INFIXFOISDIV                      { $1 }
  | ASSIGN                            { ":=" }
  | INFIXCOMP                         { $1 }
  | ANDB                              { "&&" }
  | OR                                { "||" }
  | CONS                              { "::" }
  | SEMICOLON                         { ";" }
  | LISTCONCAT                        { "@" }
  | STRINGCONCAT                      { "^" }
  | BANG                              { "!" }
  |                                   { "()" }
;

record:
/* Hack pour transformer { l x y = e, ... } en { l = fun x y -> e, ... }
   en evitant les conflits avec { l x y, ... } */
    expr EQ expr                      { appl_list_to_fun $1 $3 }
  | expr COMMA ID params EQ expr      { RCons($1,$3,param_list $4 $6) }
  | record COMMA ID params EQ expr    { RCons($1,$3,param_list $4 $6) }
  | expr COMMA ID params OVERRIDE expr
                                      { RCons(RRes($1,$3),$3,param_list $4 $6)}
  | record COMMA ID params OVERRIDE expr
                                      { RCons(RRes($1,$3),$3,param_list $4 $6)}
  |                                   { REmpty }
;

mixin:
    mixin INHERIT appl                { mixin_inherit $1 $3 }
  | mixin VAR ID params EQ expr       { mixin_var $1 $3 (param_list $4 $6) }
  | mixin CST ID params EQ expr       { mixin_cst $1 $3 (param_list $4 $6) }
  | mixin METH ID LPAR id COMMA id RPAR params EQ expr
                            { mixin_meth $1 $3 $5 $7 (param_list $9 $11) }
  | mixin METH ID LPAR id COMMA id RPAR params OVERRIDE expr
                            { mixin_meth_over $1 $3 $5 $7 (param_list $9 $11) }
  | mixin WITHOUT ID                  { mixin_without $1 $3 }
  | mixin RENAME ID AS ID             { mixin_rename $1 $3 $5 }
  |                                   { empty () }
;

params:
    id params                         { $1::$2 }
  |                                   { [] }
;

list:
    expr COMMA list                   { Appl(Appl(Ident "::", $1), $3) }
  | expr                              { Appl(Appl(Ident "::", $1), Ident "[]")}
  |                                   { Ident("[]") }
;

id:
    ID                                { ($1,None) }
  | LPAR ID COLON typexpr RPAR        { ($2,Some $4) }
;

typexpr:
    ID                                { typ_of_id $1 }
  | TVARID                            { typ_of_tvarid $1 }
  | typexpr TO typexpr                { fun_cell ($1,dvar_cell (),$3) }
  | LPAR typexpr RPAR                 { $2 }
  | LACC recordt RACC                 { record_cell (make_rec empty_cell $2) }
  | LACC RVARID COMMA recordt RACC    { record_cell
			  (make_rec (typ_of_rvarid $2 (List.map fst $4)) $4) }
  | typexpr ID                        { constr_cell ($2,$1) }
;

recordt:
    ID COLON typexpr                  { [($1,$3)] }
  | ID COLON typexpr COMMA recordt    { ($1,$3)::$5 }
  |                                   { [] }
;
