open Common;;
open Types;;
open Syntax;;
open Set;;
open Typing;;
open Value;;

let binIntInt f =
 Primitive (function Int x -> Primitive (function
                                             Int y -> Int (f x y)
                                           | _ -> failwith "binIntInt-error")
                   | _ -> failwith "binIntInt-error");;
let plus = binIntInt (+);;
let moins = binIntInt (-);;
let fois = binIntInt ( * );;
let div = binIntInt (fun x y -> if y = 0 then runtime_msg "Division by zero"
                                         else x / y);;
let modulo = binIntInt (fun x y -> if y = 0 then runtime_msg "Division by zero"
                                            else x mod y);;

let binIntBool f =
 Primitive (function Int x -> Primitive (function
                                             Int y -> Bool (f x y)
                                           | _ -> failwith "binIntBool-error")
                   | _ -> failwith "binIntBool-error");;
let inf = binIntBool (<);;
let infE = binIntBool (<=);;
let sup = binIntBool (>);;
let supE = binIntBool (>=);;
let eq = binIntBool (=);;
let diff = binIntBool (<>);;

let binBoolBool f =
 Primitive (function Bool x -> Primitive (function
                                             Bool y -> Bool (f x y)
                                           | _ -> failwith "binBoolBool-error")
                   | _ -> failwith "binBoolBool-error");;
let et = binBoolBool (&&);;
let ou = binBoolBool (||);;

let non = Primitive (function Bool b -> Bool (not b)
                            | _ -> failwith "Not-error");;

let reference = Primitive (fun v -> Location (ref v));;
let bang = Primitive (function Location l -> !l
                             | _ -> failwith "Bang-error");;
let assign =
 Primitive (function Location l -> Primitive (fun v -> l := v; Unit)
                            | _ -> failwith "Assign-error");;

let cons =
 Primitive (function v -> Primitive (function List l -> List (v::l)
                                            | _ -> failwith "Cons-error"));;
let hd =
 Primitive (function List [] -> runtime_msg "Empty list"
                   | List (v::_) -> v
                   | _ -> failwith "Head-error");;
let tl =
 Primitive (function List [] -> runtime_msg "Empty list"
                   | List (_::l) -> List l
                   | _ -> failwith "Tail-error");;
let null =
 Primitive (function List l -> Bool (l = [])
                   | _ -> failwith "Null-error");;

let listConcat =
 Primitive (function List l1 -> Primitive (function
					      List l2 -> List (l1 @ l2)
					    | _ -> failwith "ListConcat-error")
		   | _ -> failwith "ListConcat-error");;

let stringConcat =
 Primitive (function String s1 -> Primitive (function
					    String s2 -> String (s1 ^ s2)
					  | _ -> failwith "StringConcat-error")
		   | _ -> failwith "StringConcat-error");;

let seq = Primitive (function Unit -> Primitive (fun v -> v)
                            | _ -> failwith "Seq-error");;

let printint = Primitive (function Int n -> print_int n; Unit
                                 | _ -> failwith "PrintInt-error");;
let printstring = Primitive (function String s -> print_string s; Unit
                                    | _ -> failwith "PrintString-error");;
let println = Primitive (function Unit -> print_newline (); Unit
                                | _ -> failwith "PrintLn-error");;

let quit = Primitive (function Unit -> raise Quit
                             | _ -> failwith "Quit-error");;

let (start_env : env) = [
 ("true", Bool true);
 ("false", Bool false);
 ("()", Unit);
 ("+", plus);
 ("-", moins);
 ("*", fois);
 ("/", div);
 ("%", modulo);
 ("<", inf);
 ("<=", infE);
 (">", sup);
 (">=", supE);
 ("==", eq);
 ("<>", diff);
 ("&&", et);
 ("||", ou);
 ("not", non);
 ("ref", reference);
 ("!", bang);
 (":=", assign);
 ("[]", List []);
 ("::", cons);
 ("hd", hd);
 ("tl", tl);
 ("null", null);
 ("@", listConcat);
 ("^", stringConcat);
 (";", seq);
 ("print_int", printint);
 ("print_string", printstring);
 ("print_newline", println);
 ("quit", quit)
];;

let (start_envT : envT) =
 let sBool = (poly_empty, bool_cell)
 and sUnit = (poly_empty, unit_cell)
 and sIntInt () = let d = dvar_cell () in
                  gen (fun_cell (int_cell, d,
		       fun_cell (int_cell, zero_cell, int_cell))) []
 and sIntBool () = let d = dvar_cell () in
                   gen (fun_cell (int_cell, d,
			fun_cell (int_cell, zero_cell, bool_cell))) []
 and sBoolBool () = let d = dvar_cell () in
                    gen (fun_cell (bool_cell, d,
			 fun_cell (bool_cell, zero_cell, bool_cell))) []
 and sNot = (poly_empty, fun_cell (bool_cell, zero_cell, bool_cell))
 and sRef = let a = tvar_cell () and d = dvar_cell () in
            gen (fun_cell (a, d, constr_cell ("ref", a))) []
 and sBang = let a = tvar_cell () in
             gen (fun_cell (constr_cell ("ref", a), zero_cell, a)) []
 and sAssign = let a = tvar_cell () and d = dvar_cell () and d' = dvar_cell ()
               in gen (fun_cell (constr_cell ("ref", a), d,
		       fun_cell (a, d', unit_cell))) []
 and sNil = let a = tvar_cell () in gen (constr_cell ("list", a)) []
 and sCons = let a = tvar_cell () and d = dvar_cell () in
             gen (fun_cell (a, d,
		  fun_cell (constr_cell ("list", a), zero_cell,
			    constr_cell ("list", a)))) []
 and sHd = let a = tvar_cell () in
           gen (fun_cell (constr_cell ("list", a), zero_cell, a)) []
 and sTl = let a = tvar_cell () in
           gen (fun_cell (constr_cell ("list", a), zero_cell,
			  constr_cell ("list", a))) []
 and sNull = let a = tvar_cell () in
             gen (fun_cell (constr_cell ("list", a), zero_cell, bool_cell)) []
 and sListConcat = let a = tvar_cell () and d = dvar_cell () in
                   gen (fun_cell (constr_cell ("list", a), d,
			fun_cell (constr_cell ("list", a), zero_cell,
				  constr_cell ("list", a)))) []
 and sStringConcat = let a = tvar_cell () and d = dvar_cell () in
                     gen (fun_cell (string_cell, d,
			  fun_cell (string_cell, zero_cell,
		         	    string_cell))) []
 and sPrintInt = (poly_empty, fun_cell (int_cell, zero_cell, unit_cell))
 and sPrintString = (poly_empty, fun_cell (string_cell, zero_cell, unit_cell))
 and sPrintLn = (poly_empty, fun_cell (unit_cell, zero_cell, unit_cell))
 and sSeq = let a = tvar_cell () and d = dvar_cell () in
            gen (fun_cell (unit_cell, d, fun_cell (a, zero_cell, a))) []
 and sQuit = let a = tvar_cell () in
             gen (fun_cell (unit_cell, zero_cell, a)) []
 in [
 ("true", sBool);
 ("false", sBool);
 ("()", sUnit);
 ("+", sIntInt ());
 ("-", sIntInt ());
 ("*", sIntInt ());
 ("/", sIntInt ());
 ("%", sIntInt ());
 ("<", sIntBool ());
 ("<=", sIntBool ());
 (">", sIntBool ());
 (">=", sIntBool ());
 ("==", sIntBool ());
 ("<>", sIntBool ());
 ("&&", sBoolBool ());
 ("||", sBoolBool ());
 ("not", sNot);
 ("ref", sRef);
 ("!", sBang);
 (":=", sAssign);
 ("[]", sNil);
 ("::", sCons);
 ("hd", sHd);
 ("tl", sTl);
 ("null", sNull);
 ("@", sListConcat);
 ("^", sStringConcat);
 (";", sSeq);
 ("print_int", sPrintInt);
 ("print_string", sPrintString);
 ("print_newline", sPrintLn);
 ("quit", sQuit)
];;
