open Types;;
open Set;;
open Ineq;;
open Fv;;

type var_letter =
 { first : char; last : char; mutable letter : char; mutable num : int };;

let new_vl f l = { first = f; last = l; letter = f; num = 0 };;

let advance_vl vl =
 if vl.letter = vl.last
 then (vl.letter <- vl.first; vl.num <- vl.num + 1)
 else vl.letter <- char_of_int (int_of_char vl.letter + 1);;

let get_vl vl =
 String.make 1 vl.letter ^ (if vl.num = 0 then "" else string_of_int vl.num);;

let (print_constr : constr -> unit) =
 let rec loop = function
    [] -> ()
  | [lab] -> print_string lab
  | (lab::r) -> print_string lab; print_string ","; loop r
 in fun c -> print_string "{"; loop c; print_string "}";;

let (start_printing, register_poly, typ_string, row_string, deg_string) =
 let lower_case = ref (new_vl 'a' 'z')
 and upper_case = ref (new_vl 'A' 'Z')
 and typ_assoc = ref []
 and row_assoc = ref []
 and deg_assoc = ref []
 and poly_vars = ref poly_empty in
 (fun () -> lower_case := new_vl 'a' 'z';
            upper_case := new_vl 'A' 'Z';
            typ_assoc := [];
            row_assoc := [];
            deg_assoc := [];
	    poly_vars := poly_empty),
 (fun p -> poly_vars := poly_union p !poly_vars),
 (fun u -> if List.mem_assq u !typ_assoc
           then List.assq u !typ_assoc
           else let s = "'" ^ (if List.memq u !poly_vars.tvar then "" else "_")
		            ^ get_vl !lower_case in
                typ_assoc := (u,s)::(!typ_assoc);
                advance_vl !lower_case;
                s),
 (fun u -> if List.mem_assq u !row_assoc
           then List.assq u !row_assoc
           else let s = "'" ^ (if List.memq u !poly_vars.rvar then "" else "_")
		            ^ get_vl !upper_case in
                row_assoc := (u,s)::(!row_assoc);
                advance_vl !upper_case;
                s),
 (fun u -> if List.mem_assq u !deg_assoc
           then List.assq u !deg_assoc
           else let s = "^" ^ (if List.memq u !poly_vars.dvar then "" else "_")
		            ^ get_vl !lower_case in
                deg_assoc := (u,s)::(!deg_assoc);
                advance_vl !lower_case;
                s);;

let recursive_assoc t =
 let compt = ref 0 in
 let new_letter () = incr compt; "X" ^ string_of_int !compt in

 let t_assoc = ref [] in
 let processed = ref [] in

 let rec loop u =
  let u = repr u in
  match u.data with
     FunT (u1, _, u2) ->
      if List.memq u !processed
      then (if not (List.mem_assq u !t_assoc)
            then t_assoc := (u,new_letter())::(!t_assoc))
      else (processed := u::(!processed); loop u1; loop u2)
   | RecordT u1 ->
      if List.memq u !processed
      then (if not (List.mem_assq u !t_assoc)
            then t_assoc := (u,new_letter())::(!t_assoc))
      else (processed := u::(!processed); loopr u1)
   | ConstrT (_, u1) ->
      if List.memq u !processed
      then (if not (List.mem_assq u !t_assoc)
            then t_assoc := (u,new_letter())::(!t_assoc))
      else (processed := u::(!processed); loop u1)
   | _ -> ()

 and loopr u =
  let u = repr u in
  match u.data with
     LabelT (_, u1) -> loop u1
   | NodeT (u1, u2) -> loopr u1; loopr u2
   | _ -> ()

 in loop t; !t_assoc;;

let print_scheme ((p, t) : scheme) =
 let rec_tassoc = recursive_assoc t in

 let l_ineq_treated = ref setq_empty
 and l_ineq = ref setq2_empty
 and l_rvar = ref setq_empty in

 register_poly p;

 let rec loop par u =
  let u = repr u in
  if List.mem_assq u rec_tassoc
  then print_string (List.assq u rec_tassoc)
  else loop' par u

 and loop' par u =
  match u.data with
     IntT -> print_string "int"
   | BoolT -> print_string "bool"
   | UnitT -> print_string "unit"
   | StringT -> print_string "string"
   | FunT (u1,u2,u3) -> if par then print_char '('; loop true u1;
                        print_char ' '; loopd u2; print_string "-> ";
                        loop false u3; if par then print_char ')'
   | TVar -> print_string (typ_string u)
   | RecordT u1 -> loopr u1
   | ConstrT (s, u1) -> loop true u1; print_char ' '; print_string s

 and loopr u =
  let u = repr u in
  let (var,fields) = rec_fields u in
   match var with
      None -> if fields = [] then print_string "{ }"
              else (print_string "{ "; looprr fields; print_string " }")
    | Some var -> if lab_not_allowed var <> []
                  then l_rvar := setq_add var !l_rvar;
                  if fields = []
                  then print_string (row_string var)
	 	  else (print_string "{ "; print_string (row_string var);
		        print_string ", "; looprr fields; print_string " }")

 and looprr = function
    [] -> failwith "Print-scheme-looprr"
  | [(l,t)] -> print_string l; print_string " : "; loop false t
  | ((l,t)::r) -> print_string l; print_string " : "; loop false t;
                  print_string ", "; looprr r

 and loopd u =
  let u = repr u in
  match u.data with
     Zero -> print_string "^0"
   | One -> print_string "^1"
   | DVar _ -> print_string (deg_string u); treat_dvar u
   | Inf (u,v) -> loopd u; print_string "^^"; loopd v

 and treat_dvar u =
  let u = repr u in
  if not (List.memq u !l_ineq_treated)
  then let (l1,l2) = infs_sups u in
       l_ineq_treated := u::!l_ineq_treated;
       l_ineq := setq2_union (List.map (fun v -> (repr v,u)) !l1) !l_ineq;
       l_ineq := setq2_union (List.map (fun v -> (u,repr v)) !l2) !l_ineq;
       List.iter treat_dvar !l1;
       List.iter treat_dvar !l2

 in
 loop false t;
 if rec_tassoc <> [] || !l_rvar <> [] || !l_ineq <> []
 then begin
       print_newline (); print_string "where ";
       let foo (u,s) =
        print_string s; print_string " = "; loop' false u; print_newline ()
       in List.iter foo rec_tassoc;
       let foo2 var = print_string (row_string var); print_string " :: ";
                      print_constr (lab_not_allowed var); print_char ' '
       in List.iter foo2 !l_rvar;
       let foo3 (d1, d2) = print_string (deg_string d1); print_string "<=";
                           print_string (deg_string d2); print_char ' '
       in List.iter foo3 !l_ineq
      end;;

let print_typ t = print_scheme (fv_typ t, t);;
