open Common;;
open Types;;
open Set;;
open Syntax;;
open Ineq;;
open Fv;;
open Degree;;
open Tprinter;;

let (gen : typ cell -> envT -> scheme) = fun t e ->
 let s_gen = poly_minus (fv_typ t) (fv_envT e) in
 let s_deg_impl = setq_minus s_gen.dvar (fv2_deg t) in
 List.iter suppress s_deg_impl;
 ({ s_gen with dvar = setq_minus s_gen.dvar s_deg_impl }, t);;

let generalize t e m =
 let rec loop e = function
    Appl _ -> false
  | Let (binds,n) -> let e' = (List.map (fun (x,m) -> (x,loop e m)) binds)
                              @ e in
                     loop e' n
  | LetRec (binds,n) -> let e' = (List.map (fun (x,_) -> (x,false)) binds)
                                 @ e in
                        let e'' = (List.map (fun (x,m) -> (x,loop e' m)) binds)
			          @ e in
  			loop e'' n
  | Ident x -> if List.mem_assoc x e then List.assoc x e else true
  | Fun _ -> true
  | IntE _ -> true
  | StringE _ -> true
  | If _ -> false
  | REmpty -> true
  | RCons (m,l,n) -> loop e m && loop e n
  | RSel _ -> false
  | RRes _ -> false
  | TypeConstr (m,_) -> loop e m
 in if loop [] m then gen t e else (poly_empty, t);;

let instance ((p,t) : scheme) =
(* if p = poly_empty then t else *)
 begin
  let tla = List.map (fun u -> (u, tvar_cell ())) p.tvar
  and rla = List.map (fun u -> (u, rvar_cell (lab_not_allowed u))) p.rvar
  and dla = List.map (fun u -> (u, dvar_cell ())) p.dvar in

  let tassoc u = if List.mem_assq u tla then List.assq u tla else u
  and rassoc u = if List.mem_assq u rla then List.assq u rla else u
  and dassoc u = if List.mem_assq u dla then List.assq u dla else u in

  let foo (u,v) =
   let u = repr u in
   let (l1,l2) = infs_sups u
   and (l1',l2') = infs_sups v in
   l1' := List.map dassoc !l1;
   l2' := List.map dassoc !l2
  in List.iter foo dla;

  let processed_tvar = ref []
  and processed_rvar = ref []
  and processed_dvar = ref [] in

  let rec loop u =
   let u = repr u in
   if List.mem_assq u !processed_tvar then List.assq u !processed_tvar
   else begin
    let t = tvar_cell () in
    processed_tvar := (u,t)::(!processed_tvar);
    let v = match u.data with
       TVar -> tassoc u
     | FunT (u1, u2, u3) -> fun_cell (loop u1, loopd u2, loop u3)
     | RecordT u1 -> record_cell (loopr u1)
     | ConstrT (l, u1) -> constr_cell (l, loop u1)
     | _ -> u
    in t.repr <- v; v
   end

  and loopr u =
   let u = repr u in
   if List.mem_assq u !processed_rvar then List.assq u !processed_rvar
   else begin
    let r = rvar_cell [] in
    processed_rvar := (u,r)::(!processed_rvar);
    let v = match u.data with
       RVar _ -> rassoc u
     | LabelT (l, u1) -> label_cell (l, loop u1)
     | NodeT (u1, u2) -> node_cell (loopr u1, loopr u2)
     | _ -> u
    in r.repr <- v; v
   end

  and loopd u =
   let u = repr u in
   if List.mem_assq u !processed_dvar then List.assq u !processed_dvar
   else begin
    let d = dvar_cell () in
    processed_dvar := (u,d)::(!processed_dvar);
    let v = match u.data with
       DVar _ -> dassoc u
     | Inf (u1, u2) -> inf_cell (loopd u1, loopd u2)
     | _ -> u
    in d.repr <- v; v
   end

  in loop t
 end;;

let mgu_fields l1 l2 =
 let rec loop aux1 aux2 auxc l1 l2 = match (l1,l2) with
    ([],l2) -> (List.rev aux1 @ l2, List.rev aux2, List.rev auxc)
  | (l1,[]) -> (List.rev aux1, List.rev aux2 @ l1, List.rev auxc)
  | ((lab1,t1)::r1,(lab2,t2)::r2) ->
       if lab1 < lab2 then loop aux1 ((lab1,t1)::aux2) auxc r1 l2
       else if lab1 = lab2 then loop aux1 aux2 ((t1,t2)::auxc) r1 r2
       else loop ((lab2,t2)::aux1) aux2 auxc l1 r2
 in loop [] [] [] l1 l2;;

let check_rec u fields =
 let l = lab_not_allowed u in
 let foo b (lab,_) = b && not (List.mem lab l) in
 List.fold_left foo true fields;;

let rec mgu_degree_one u =
 let u = repr u in
 match u.data with
   One -> u.repr <- one_cell
 | Zero -> type_runtime_msg "Unable to satisfy 0 = 1"
 | DVar (l1,l2) -> List.iter (suppress_ineq u) !l1;
                   u.repr <- one_cell;
                   List.iter mgu_degree_one !l2
 | Inf (u1, u2) -> mgu_degree_one u1;
                   mgu_degree_one u2;;

let rec mgu_degree_zero u =
 let u = repr u in
 match u.data with
   Zero -> u.repr <- zero_cell
 | One -> type_runtime_msg "Unable to satisfy 0 = 1"
 | DVar (l1,l2) -> List.iter (suppress_ineq u) !l2;
                   u.repr <- zero_cell;
                   List.iter mgu_degree_zero !l1
 | Inf _ -> failwith "Mgu_degree_zero-Error";;

(* en fait toujours appele sous la forme "add_ineq (DVar _) v" *)
let rec add_ineq u v =
 let u = repr u and v = repr v in
 match (u.data, v.data) with
    (One, Zero) -> type_runtime_msg "Constraint 1 <= 0"
  | (Zero, _) | (_, One) -> ()
  | (DVar _, Zero) -> mgu_degree_zero u
  | (One, DVar _) -> mgu_degree_one v
  | (DVar _, DVar _) when u == v -> ()
  | (DVar (_,l1), DVar (l2,_)) -> l1 := setq_add v !l1;
                                  l2 := setq_add u !l2
  | (_, Inf (u1,u2)) -> add_ineq u u1; add_ineq u u2
  | _ -> failwith "Add_ineq-Error";;

let mgu_degree u v =
 let u = repr u and v = repr v in
 if u == v then () else
 match (u.data,v.data) with
    (Zero, _) -> mgu_degree_zero v
  | (One, _) -> mgu_degree_one v
  | (_, Zero) -> mgu_degree_zero u
  | (_, One) -> mgu_degree_one u
  | (DVar (l1,l2), DVar (l1',l2')) ->
      u.repr <- v;
      l1' := setq_minus (setq_union !l1 !l1') [u; v];
      l2' := setq_minus (setq_union !l2 !l2') [u; v];
      List.iter (replace_up u v) !l1;
      List.iter (replace_down u v) !l2
  | _ -> failwith "Mgu_degree-Error";;

let rec mgu_typ u v =
 let u = repr u and v = repr v in
 if u == v then () else
 match (u.data,v.data) with
    (IntT, IntT) | (BoolT, BoolT) | (UnitT, UnitT) | (StringT, StringT) ->
       u.repr <- v
  | (FunT (u1,u2,u3), FunT (v1,v2,v3)) ->
      u.repr <- v; mgu_typ u1 v1; mgu_degree u2 v2; mgu_typ u3 v3
  | (RecordT u1, RecordT v1) -> u.repr <- v; mgu_rec u1 v1
  | (ConstrT (s, u1), ConstrT (s', v1)) when s = s' ->
      u.repr <- v; mgu_typ u1 v1
  | (TVar, _) -> u.repr <- v
  | (_, TVar) -> v.repr <- u
  | _ -> type_runtime (fun () -> print_string "Unable to unify "; print_typ u;
                       print_newline (); print_string "and "; print_typ v;
                       print_newline ())

and mgu_rec u v =
 let u = repr u and v = repr v in
 if u == v then () else
 let err () = type_runtime (fun () -> print_string "Unable to unify ";
			              print_typ (record_cell u);
				      print_newline ();
				      print_string "and ";
				      print_typ (record_cell v);
                                      print_newline ()) in
 let (u_var, u_fields) = rec_fields u
 and (v_var, v_fields) = rec_fields v in
 let (u_bind, v_bind, common) = mgu_fields u_fields v_fields in
 begin
  match (u_var, v_var) with
     (None, None) -> if u_bind = [] && v_bind = [] then () else err ()
   | (None, Some var2) ->
       if u_bind = [] && check_rec var2 v_bind
       then (var2.repr <- make_rec empty_cell v_bind)
       else err ()
   | (Some var1, None) ->
       if v_bind = [] && check_rec var1 u_bind
       then (var1.repr <- make_rec empty_cell u_bind)
       else err ()
   | (Some var1, Some var2) ->
       if check_rec var1 u_bind && check_rec var2 v_bind
       then (let new_lab = set_union (lab_not_allowed var1)
	 	                     (lab_not_allowed var2) in
             let nc = rvar_cell new_lab in
             var1.repr <- make_rec nc u_bind;
             var2.repr <- make_rec nc v_bind)
       else err ()
 end;
 List.iter (fun (t1, t2) -> mgu_typ t1 t2) common;;

let rec infer (e : envT) m =
 try match m with
   Appl (m, n) -> let (t1, f1) = infer e m in
                  let (t2, f2) = infer e n in
		  let a = tvar_cell () and p = dvar_cell () in
		  mgu_typ t1 (fun_cell (t2, p, a));
		  let g = fonc_of_var zero_cell m ^^ appl_fonc p f2 m n in
		  (a, g)
 | Let (binds, n) -> let (xs, ms) = List.split binds in
                     let (ts, fs) = List.split (List.map (infer e) ms) in
		     let ts' = List.map2 (fun t m -> generalize t e m) ts ms in
		     let e' = (List.combine xs ts') @ e in
		     let (t2, f2) = infer e' n in
		     let g = (List.fold_left (^^) one fs) ^^ f2 ^^
		       (if is_pure n then one
		        else List.fold_left2
 			       (fun d x m -> d ^^ fonc_of_var (f2 x) m)
			       one xs ms)
		     in (t2, g)
 | LetRec (binds, n)  -> let (xs, ms) = List.split binds in
			 let bs = List.map (fun _ -> tvar_cell ()) xs in
			 let e' = (List.map2 (fun x b -> (x, (poly_empty, b)))
			           xs bs) @ e in
			 let (ts, fs) = List.split (List.map (infer e') ms) in
			 List.iter2 mgu_typ ts bs;
			 let ts' = List.map2 (fun t m -> generalize t e m)
			                     ts ms in
			 let e'' = (List.combine xs ts') @ e in
			 let (t2, f2) = infer e'' n in
			 List.iter (fun x ->
			   List.iter (fun f -> mgu_degree_one (f x)) fs) xs;
			 let g = (List.fold_left (^^) one fs) ^^ f2 ^^
			   (if is_pure n then one
 			    else List.fold_left2
 			           (fun d x m -> d ^^ fonc_of_var (f2 x) m)
			           one xs ms)
			 in (t2, g)
 | Fun (x, topt, m)  -> let a = match topt with Some t -> t
                                              | None -> tvar_cell () in
			let (t, f) = infer ((x, (poly_empty, a))::e) m in
			let p = dvar_cell () in
			add_ineq p (f x);
			(fun_cell (a, p, t), one)
 | Ident x  -> (instance (search_env x e), zero_x x)
 | IntE _  -> (int_cell, one)
 | StringE _ -> (string_cell, one)
 | If (b, m, n)  -> let (t1, f1) = infer e b in
                    mgu_typ t1 bool_cell;
                    let (t2, f2) = infer e n in
		    let (t3, f3) = infer e m in
		    mgu_typ t3 t2;
		    let g = f1 ^^ f2 ^^ f3 in
		    (t3, g)
 | REmpty -> (record_cell empty_cell, one)
 | RCons (m, l, n) -> let (t1, f1) = infer e m in
                      let a = rvar_cell [l] in
		      mgu_typ t1 (record_cell a);
		      let (t2, f2) = infer e n in
		      let g = f1 ^^ f2 in
		      let r = make_rec a [(l, t2)] in
		      (record_cell r, g)
 | RSel (m, l) -> let (t, f) = infer e m in
                  let a = rvar_cell [l] and b = tvar_cell () in
		  let r = make_rec a [(l, b)] in
		  mgu_typ t (record_cell r);
		  (b, f)
 | RRes (m, l) -> let (t, f) = infer e m in
                  let a = rvar_cell [l] and b = tvar_cell () in
		  let r = make_rec a [(l, b)] in
		  mgu_typ t (record_cell r);
		  (record_cell a, f)
 | TypeConstr (m, t) -> let (t1, f) = infer e m in
                        mgu_typ t t1;
                        (t1, f)
 with Type_runtime f -> runtime (fun () -> f ();
                                 if !verbose_errors then
                                 begin
                                  print_string "while typing ";
                                  print_expr false m;
                                  print_newline ()
                                 end);;
