open Common;;
open Types;;
open Syntax;;

exception Quit;;

type value =
   Closure of ident * expr * env
 | RecVal of recval
 | Unit
 | Int of int
 | Bool of bool
 | String of string
 | Location of value ref
 | Record of record
 | List of value list
 | Primitive of (value -> value)
and recval = value option ref
and env = (ident * value) list
and record = (label * value) list;;

let search_env id (env : env) =
 try
  List.assoc id env
 with Not_found -> runtime_msg ("Identifier " ^ id ^ " not defined");;

let sel_record (l : label) (r : record) = List.assoc l r;;

let (supp_record : label -> record -> record) = List.remove_assoc;;

let ins_record (l : label) v =
 let rec (loop : record->record) = function
    [] -> [(l,v)]
  | ((l',v')::r) -> if l < l' then (l,v)::(l',v')::r
                              else (l',v')::(loop r)
 in loop;;

let rec print_value = function
   Closure _ -> print_string "<fun>"
 | RecVal r -> begin
                match !r with
                   None -> failwith "RecVal-error"
                 | Some v -> print_value v
               end
 | Unit -> print_string "()"
 | Int i -> print_int i
 | String s -> print_string "\""; print_string s; print_string "\""
 | Bool b -> print_string (if b then "true" else "false")
 | Location l -> print_string "ref "; print_value !l
 | Record r -> print_string "{ "; print_record r; print_string "}"
 | List l -> print_string "["; print_list l; print_string "]"
 | Primitive op -> print_string "<primitive>"

and print_record = function
   [] -> ()
 | [(l,v)] -> print_string l; print_string " = "; print_value v;
              print_string " "
 | ((l,v)::r) -> print_string l; print_string " = "; print_value v;
                 print_string ", "; print_record r

and print_list = function
   []  -> ()
 | [v] -> print_value v
 | (v::r) -> print_value v; print_string ", "; print_list r;;
